@@ -8,6 +8,10 @@ from types import FunctionType, MethodType
 import inspect
 import traceback
 
+def is_torch(model):
+    # Method Resolution Order by C3 Linearization algorithm
+    return any(str(t) == "<class 'torch.nn.modules.module.Module'>" for t in reversed(type(model).__mro__))
+
 def availableGPU():
 # pip install nvidia-ml-py3
     try:
@@ -49,6 +53,11 @@ def availableGPU():
     print('ids =', ids)
     device_id = [*ids][random.randrange(0, len(ids))]
     print('selected device_id =', device_id)
+    import os
+    gpu_count = os.environ.get('gpu_count')
+    if gpu_count:
+        ...
+        
     return device_id
 
 def initialize_vocab(file, start=2):
@@ -174,11 +183,9 @@ class KerasModel:
                 self.model.eval()
                 import torch
                 if torch.cuda.is_available():
-                    device = availableGPU()
-                    if device >= 0:
-                        self.model.to(device)
-                        self.device = device
-                        return
+                    self.model.to(0)
+                    self.device = 0
+                    return
 
         self.device = -1
 
@@ -523,30 +530,28 @@ class KerasModel:
     def save_weights(self):
         modelFile = self.modelFile
         print('\nsaving model to', modelFile)
+        import os
+        from std.file import createNewPath
+        createNewPath(os.path.dirname(modelFile))
+
         if modelFile.endswith('.h5'):
             import h5py
-            try:
-                with h5py.File(modelFile, 'w') as f:
-                    self.iostream(self.model, f, mode='w')
-            except FileNotFoundError:
-                import os
-                std.file.createNewPath(os.path.dirname(modelFile))
-                with h5py.File(modelFile, 'w') as f:
-                    self.iostream(self.model, f, mode='w')
+            with h5py.File(modelFile, 'w') as f:
+                self.iostream(self.model, f, mode='w')
+
         elif modelFile.endswith('.pt'):
             import torch
-            try:
-                torch.save(self.model, modelFile)
-            except FileNotFoundError:
-                import os
-                std.file.createNewPath(os.path.dirname(modelFile))
-                torch.save(self.model, modelFile)
+            torch.save(self.model, modelFile)
 
+        elif modelFile.endswith('.bin'):
+            import torch
+            torch.save(self.model.state_dict(), modelFile)
+                
     @computed
     def is_torch(self):
-        return str(type(self.model).__mro__[-2]) == "<class 'torch.nn.modules.module.Module'>"
+        return is_torch(self.model)
+
 
-            
 def get_tensor_shape(tensor, axis=None):
     if axis is None:
         shape = []
@@ -651,6 +656,14 @@ class Sequence:
     def __getitem__(self, index):
         return self.arr[index]
     
+    def get_memory_allocation(self, tensor):
+        shape = get_tensor_shape(tensor)
+        if isinstance(self.dynamic_size, (FunctionType, MethodType)):
+            return self.dynamic_size(*shape)
+
+        from _functools import reduce
+        return reduce(lambda x, y: x * y, shape)
+        
     def batches(self): 
         if self.dynamic_size:
             if isinstance(self.x_name, str):
@@ -658,35 +671,31 @@ class Sequence:
             else:
                 x_name = self.x_name[0]
                 
-            from _functools import reduce
-            seq_length = 0
+            memorySize = 0
             for inst in self.training_list:
-                shape = get_tensor_shape(getattr(inst, x_name))
+                memorySize += self.get_memory_allocation(getattr(inst, x_name))
 
-                if isinstance(self.dynamic_size, (FunctionType, MethodType)):
-                    seq_length += self.dynamic_size(*shape)
-                else:
-                    seq_length += reduce(lambda x, y: x * y, shape)
-
-            seq_length /= len(self.training_list)
-            average_memory = self.batch_size * seq_length
-            i = 0
+            memorySize /= len(self.training_list)
+            batch_memory = self.batch_size * memorySize
+            
             batch_size = self.batch_size
+            _memorySize = self.get_memory_allocation(getattr(self.training_list[batch_size - 1], x_name))
+            if _memorySize < memorySize:
+                _batch_size = max(1, int(batch_memory / _memorySize))
+                if _batch_size > batch_size:
+                    batch_size = _batch_size
+                    memorySize = _memorySize
+                    print('initial batch_size =', batch_size)
+
+            i = 0
             while i < len(self.training_list):
-                batch = self.training_list[i:i + batch_size]
-                shape = get_tensor_shape(getattr(batch[-1], x_name))
-                
-                if isinstance(self.dynamic_size, (FunctionType, MethodType)):
-                    _seq_length = self.dynamic_size(*shape)
-                else:
-                    _seq_length = reduce(lambda x, y: x * y, shape)
-                
-                if _seq_length > seq_length:
-                    seq_length = _seq_length
-                    _batch_size = max(1, int(average_memory / seq_length))
+                _memorySize = self.get_memory_allocation(getattr(self.training_list[i + batch_size - 1], x_name))
+                if _memorySize > memorySize:
+                    _batch_size = max(1, int(batch_memory / _memorySize))
                     if _batch_size < batch_size:
                         batch_size = _batch_size
-                        print('adjust batch_size to', batch_size, 'for', shape)
+                        memorySize = _memorySize
+                        print('adjust batch_size to', batch_size)
      
                 yield self.training_list[i:i + batch_size]
                 i += batch_size
@@ -1159,4 +1168,4 @@ class StepRunner:
 
     @computed
     def is_torch(self):
-        return str(type(self.model).__mro__[-2]) == "<class 'torch.nn.modules.module.Module'>"
+        return is_torch(self.model)
